import re
from sdetools.extlib.defusedxml.ElementTree import fromstring, ParseError
from xml.etree import ElementTree as etree

from sdetools.sdelib.cmd import BaseCommand
from sdetools.sdelib.commons import UsageError
from sdetools.sdelib.restclient import APICallError
from sdetools.sdelib.sdeapi import APIBase
from sdetools.sdelib import log_mgr


logger = log_mgr.mods.add_mod(__name__)


class Command(BaseCommand):
    """
    sdetools command plugin for RSAM integration.

    The purpose of this command is to take in XML input (coming from RSAM)
    and perform actions on the given user based on the data provided.
    """

    help = "Command for user management that integrates with RSAM"
    name = "rsam"

    ATTRIBUTE_MAP_KEY = "provAttributeMap"
    RESPONSE_ROOT_KEY = "provisionResponse"

    ROLE_TO_GROUP_MAPPING = {
        'TMASC': 'Project Lead',
        'User': 'User',
        'Administrator': 'Admin',
    }

    class Messages:
        GROUP_NOT_FOUND = "SDE group not found for corresponding LOB and Role Name."
        INVALID_ACTION = "Invalid action."
        INVALID_KEY = "Value for {0} cannot be empty."
        INVALID_ROLE = "Invalid role name provided."
        INVALID_XML = "Invalid XML data."
        MISSING_KEY = "Invalid input. Missing key: {0}"
        ROLE_NOT_FOUND = "SDE role not found."
        USER_NOT_FOUND = "No user found with the provided email."
        USER_ADDED = "User added."
        USER_UPDATED = "User updated."
        USER_REMOVED_FROM_GROUP = "User has been removed from requested group."
        USER_DEACTIVATED = "User has been removed from all groups and deactivated."

    def configure(self):
        self.api = APIBase(self.config)
        self.config.opts.add('raw_data', "Raw XML data")
        self.config.opts.add('data', "Parsed XML data")

        self.status_code = 200
        self.message = ""
        self.namespace = None

    def _find_element(self, namespace, parent, key):
        """
        Find element with the given key in an XML tree under
        the given namespace and parent node.
        """
        key = '{' + namespace + '}' + key
        return parent.find(key)

    def _parse_data(self, xml_data):
        """
        Parse given RSAM XML input to a dict of user attributes.

        From the incoming data, we are only interested in a subset of the XML.
        An example of the subset is given below:

        <provAttributeMap>
           <entry>
               <key>ACTION</key>
               <value>Add</value>
           </entry>
           <entry>
               <key>SID</key>
               <value>I191727</value>
           </entry>
           <entry>
               <key>First Name</key>
               <value>Sathish</value>
           </entry>
            <entry>
               <key> Last Name </key>
               <value>Swaminathan</value>
           </entry>
            <entry>
               <key> Email</key>
               <value>Sathish.swaminathan@xyz.com</value>
           </entry>
            <entry>
               <key> LOB </key>
               <value>CIB</value>
           </entry>
            <entry>
               <key> RoleName</key>
               <value>Administrator</value>
           </entry>
        </provAttributeMap>
        """
        import pdb
        pdb.set_trace()

        root = fromstring(xml_data)

        # extract namespace
        match = re.search(r"\{([A-Za-z0-9_://.]+)\}", root.tag)

        if match:
            self.namespace = match.group(1)

        if not self.namespace:
            raise AssertionError("Missing namespace.")

        attribute_map = self._find_element(self.namespace, root, Command.ATTRIBUTE_MAP_KEY)

        attributes = {}

        for child in attribute_map.getchildren():
            key_tag = self._find_element(self.namespace, child, 'key')
            value_tag = self._find_element(self.namespace, child, 'value')

            key = key_tag.text or ''
            key = key.strip().lower()
            value = value_tag.text or ''

            attributes[key] = value.strip()

        return attributes

    def get_sde_role(self, LOB, role_name):
        """
        Get corresponding SDE role id for given RSAM Line of Business
        and role name.

        From the XML input we get LOB and role name. The corresponding role
        role name in SDE is a combination of the LOB + role name except for
        the case where the role name is 'Administrator'.

        e.g.
        LOB: CIB. Role: TMASC -> SDE role = CIB TMASC
        LOB: CIB. Role: Administrator -> SDE role = Administrator
        """
        if not role_name:
            return None

        sde_role_name = '%s %s' % (LOB, role_name)

        if role_name == 'Administrator':
            sde_role_name = role_name

        role = next((r for r in self.sde_roles
                     if r['name'] == sde_role_name), None)

        if not role:
            raise AssertionError(Command.Messages.ROLE_NOT_FOUND)

        return role['id']

    def get_sde_group(self, LOB, role_name):
        """
        Get corresponding SDE group id for given RSAM Line of Business
        and role name.

        From the XML input we get LOB and role name. The corresponding role
        group name in SDE is calculated as follows.

        A matching value of the given role name is looked up in the following
        mapping:

        'TMASC': 'Project Lead'
        'User': 'User'
        'Administrator': 'Admin'

        This value is appended to the LOB to get the SDE group name, unless
        the value is 'Admin', in which case the group name 'TM Admin' is used.

        e.g.
        LOB: CIB. Role: TMASC -> SDE group = CIB Project Lead
        LOB: CIB. Role: Administrator -> SDE group = TM Admin
        """
        if not role_name:
            return None

        try:
            group_name = Command.ROLE_TO_GROUP_MAPPING[role_name]
        except KeyError:
            raise AssertionError(Command.Messages.INVALID_ROLE)

        sde_group_name = '%s %s' % (LOB, group_name)

        if group_name == 'Admin':
            sde_group_name = 'TM Admin'

        group = next((g for g in self.sde_groups
                      if g['name'] == sde_group_name), None)

        if not group:
            raise AssertionError(Command.Messages.GROUP_NOT_FOUND)

        return group['id']

    def assert_user_exists(self, user):
        if not user:
            raise AssertionError(Command.Messages.USER_NOT_FOUND)

    def assert_group_exists(self, group):
        if not group:
            raise AssertionError(Command.Messages.GROUP_NOT_FOUND)

    def process_add(self, existing_user, pay_load, target_group):
        """Create a new user if user doesn't exist, else update the user."""
        self.assert_group_exists(target_group)

        if not existing_user:
            logger.info("User doesn't exist. Creating new user")

            pay_load['groups'] = [target_group]
            self.api.create_user(**pay_load)
            self.message = Command.Messages.USER_ADDED

        else:
            logger.info("User already exists, attempting to update user details")

            current_groups = [g['id'] for g in existing_user['groups']]
            new_groups = current_groups + [target_group]

            pay_load['groups'] = new_groups

            self.api.update_user(existing_user['id'], **pay_load)
            self.message = Command.Messages.USER_UPDATED

    def process_remove(self, existing_user, pay_load, target_group):
        """Remove the user from the provided group."""
        self.assert_group_exists(target_group)

        current_groups = [g['id'] for g in existing_user['groups']]
        new_groups = [g for g in current_groups if g != target_group]

        pay_load['groups'] = new_groups

        self.api.update_user(existing_user['id'], **pay_load)
        self.message = Command.Messages.USER_REMOVED_FROM_GROUP

    def process_remove_all(self, existing_user, pay_load):
        """Remove the user from all groups and mark the user inactive."""
        pay_load.update({
            'groups': [],
            'is_active': False,
            'role': existing_user['role']['id']
        })

        self.api.update_user(existing_user['id'], **pay_load)
        self.message = Command.Messages.USER_DEACTIVATED

    def process_request(self):
        xml_data = self.config.get('raw_data', None)

        if not isinstance(xml_data, basestring):
            raise AssertionError(Command.Messages.INVALID_XML)

        self.sde_users = self.api.get_users()
        self.sde_groups = self.api.get_groups()
        self.sde_roles = self.api.get_global_roles()

        request_data = self._parse_data(xml_data)

        required_keys = ['action', 'email']

        for key in required_keys:
            try:
                if not request_data[key]:
                    raise AssertionError(Command.Messages.INVALID_KEY.format(key))
            except KeyError:
                raise AssertionError(Command.Messages.MISSING_KEY.format(key))

        action = request_data['action'].lower()
        email = request_data['email']
        role_name = request_data.get('rolename', '')
        LOB = request_data.get('lob', '')

        sde_group = self.get_sde_group(LOB, role_name)
        sde_role = self.get_sde_role(LOB, role_name)

        existing_user = next((u for u in self.sde_users
                              if u['email'] == email), None)

        pay_load = {
            'email': email,
            'role': sde_role,
        }

        if action == "add":

            try:
                pay_load.update({
                    'first_name': request_data['first name'],
                    'last_name': request_data['last name'],

                })
            except KeyError, e:
                raise AssertionError(Command.Messages.MISSING_KEY.format(e.message))

            self.process_add(existing_user, pay_load, target_group=sde_group)

        elif action == "remove":
            self.assert_user_exists(existing_user)

            pay_load.update({
                'first_name': existing_user['first_name'],
                'last_name': existing_user['last_name']
            })

            if not sde_group:
                return self.process_remove_all(existing_user, pay_load)

            self.process_remove(existing_user, pay_load, target_group=sde_group)

        else:
            raise AssertionError("Invalid action.")

    def render_response(self, status_code, message):
        """Create an XML response comprising of a message and a status code"""
        namepsace = {}

        if self.namespace:
            namepsace['xmlns'] = unicode(self.namespace)

        # Root element
        root = etree.Element(Command.RESPONSE_ROOT_KEY, **namepsace)

        # Status code element
        status_code_node = etree.SubElement(root, 'statusCode')
        status_code_node.text = str(status_code)

        # Message element
        message_node_title = 'status' if status_code == 200 else 'errorMessage'
        message_node = etree.SubElement(root, message_node_title)
        message_node.text = message

        response = etree.tostring(root, encoding='UTF-8')

        return response

    def handle(self):
        try:
            self.process_request()

        except AssertionError, e:
            self.status_code = 400
            self.message = e.message

        except ParseError:
            self.status_code = 400
            self.message = Command.Messages.INVALID_XML

        except Exception:
            self.status_code = 500
            self.message = "An error occurred while running the command."

        finally:
            response = self.render_response(self.status_code, self.message)
            self.emit.queue(return_value=str(response))

            return True
